@@ -10,9 +10,10 @@ from . import models
 class SyncJobAdmin(admin.ModelAdmin):
     list_display = (
         'id', 'start_time', 'end_time', 'duration_or_zero', 'entity_name',
-        'success', 'added', 'updated', 'deleted', 'sync_type',
+        'synchronizer_class', 'success', 'added', 'updated', 'deleted',
+        'sync_type',
     )
-    list_filter = ('sync_type', 'success', 'entity_name', )
+    list_filter = ('sync_type', 'success', 'entity_name', 'synchronizer_class')
 
     def duration_or_zero(self, obj):
         """
