@@ -1,6 +1,8 @@
 import logging
 import re
 import json
+import pytz
+from urllib.parse import urlparse
 
 import requests
 from retrying import retry
@@ -9,6 +11,32 @@ from django.conf import settings
 from django.core.cache import cache
 from djconnectwise.utils import DjconnectwiseSettings
 
+# Cloud URLs:
+# https://developer.connectwise.com/Products/Manage/Developer_Guide#Cloud_URLs
+CW_CLOUD_URLS = {
+    'au.myconnectwise.net': 'api-au.myconnectwise.net',
+    'eu.myconnectwise.net': 'api-eu.myconnectwise.net',
+    'na.myconnectwise.net': 'api-na.myconnectwise.net',
+    'aus.myconnectwise.net': 'api-aus.myconnectwise.net',
+    'za.myconnectwise.net': 'api-za.myconnectwise.net',
+}
+COMPANY_INFO_REQUIRED = 'company-info-required'
+CW_CLOUD_DOMAIN = 'myconnectwise.net'
+DEFAULT_CW_API_CODEBASE = 'v4_6_release/'
+
+CW_RESPONSE_MAX_RECORDS = 1000  # The greatest number of records ConnectWise
+# will send us in one response.
+RETRY_WAIT_EXPONENTIAL_MULTAPPLIER = 1000  # Initial number of milliseconds to
+# wait before retrying a request.
+RETRY_WAIT_EXPONENTIAL_MAX = 10000  # Maximum number of milliseconds to wait
+# before retrying a request.
+CW_DEFAULT_PAGE = 1  # CW Pagination is 1-indexed
+CONTENT_DISPOSITION_RE = re.compile(
+    '^attachment; filename=\"{0,1}(.*?)\"{0,1}$'
+)
+
+logger = logging.getLogger(__name__)
+
 
 class ConnectWiseAPIError(Exception):
     """Raise this, not request exceptions."""
@@ -37,24 +65,6 @@ class ConnectWiseRecordNotFoundError(ConnectWiseAPIClientError):
     pass
 
 
-COMPANY_INFO_REQUIRED = 'company-info-required'
-CW_CLOUD_DOMAIN = 'myconnectwise.net'
-DEFAULT_CW_API_CODEBASE = 'v4_6_release/'
-
-CW_RESPONSE_MAX_RECORDS = 1000  # The greatest number of records ConnectWise
-# will send us in one response.
-RETRY_WAIT_EXPONENTIAL_MULTAPPLIER = 1000  # Initial number of milliseconds to
-# wait before retrying a request.
-RETRY_WAIT_EXPONENTIAL_MAX = 10000  # Maximum number of milliseconds to wait
-# before retrying a request.
-CW_DEFAULT_PAGE = 1  # CW Pagination is 1-indexed
-CONTENT_DISPOSITION_RE = re.compile(
-    '^attachment; filename=\"{0,1}(.*?)\"{0,1}$'
-)
-
-logger = logging.getLogger(__name__)
-
-
 class CompanyInfoManager:
     COMPANYINFO_ENDPOINT = '{}/login/companyinfo/{}'
 
@@ -141,6 +151,7 @@ class ConnectWiseAPIClient(object):
         self,
         company_id=None,
         server_url=None,
+        client_id=None,
         api_public_key=None,
         api_private_key=None,
     ):
@@ -148,6 +159,8 @@ class ConnectWiseAPIClient(object):
             company_id = settings.CONNECTWISE_CREDENTIALS['company_id']
         if not server_url:
             server_url = settings.CONNECTWISE_SERVER_URL
+        if not client_id:
+            client_id = settings.CONNECTWISE_CLIENTID
         if not api_public_key:
             api_public_key = settings.CONNECTWISE_CREDENTIALS['api_public_key']
         if not api_private_key:
@@ -162,7 +175,8 @@ class ConnectWiseAPIClient(object):
         self.company_id = company_id
         self.api_public_key = api_public_key
         self.api_private_key = api_private_key
-        self.server_url = server_url
+        self.server_url = self.change_cw_cloud_url(server_url)
+        self.client_id = client_id
         self.auth = (
             '{0}+{1}'.format(company_id, self.api_public_key),
             '{0}'.format(self.api_private_key),
@@ -197,7 +211,9 @@ class ConnectWiseAPIClient(object):
             primary_error_msg = '{}.'.format(stripped_message)
             if error.get('errors'):
                 for error_message in error.get('errors'):
-                    messages.append('{}.'.format(error_message.get('message')))
+                    messages.append(
+                        '{}.'.format(error_message.get('message').rstrip('.'))
+                    )
 
             messages = ' The error was: '.join(messages)
 
@@ -227,6 +243,20 @@ class ConnectWiseAPIClient(object):
 
         return codebase_updated
 
+    def get_headers(self):
+        headers = {}
+
+        response_version = self.request_settings.get('response_version')
+        if response_version:
+            headers['Accept'] = \
+                'application/vnd.connectwise.com+json; version={}' \
+                .format(response_version)
+
+        if self.client_id:
+            headers['clientId'] = self.client_id
+
+        return headers
+
     def fetch_resource(self, endpoint_url, params=None, should_page=False,
                        retry_counter=None,
                        *args, **kwargs):
@@ -278,7 +308,8 @@ class ConnectWiseAPIClient(object):
                 response = requests.get(
                     endpoint,
                     auth=self.auth,
-                    timeout=self.timeout
+                    timeout=self.timeout,
+                    headers=self.get_headers(),
                 )
                 logger.info(" URL: {}".format(response.url))
 
@@ -351,6 +382,7 @@ class ConnectWiseAPIClient(object):
                 json=body,
                 auth=self.auth,
                 timeout=self.timeout,
+                headers=self.get_headers(),
             )
         except requests.RequestException as e:
             logger.error(
@@ -378,24 +410,21 @@ class ConnectWiseAPIClient(object):
             self._log_failed(response)
             raise ConnectWiseAPIError(response)
 
+    def change_cw_cloud_url(self, server_url):
+        """
+        Replace the user-facing CW CLoud URLs with the API URLs.
 
-class ProjectAPIClient(ConnectWiseAPIClient):
-    API = 'project'
-    ENDPOINT_PROJECTS = 'projects/'
-    ENDPOINT_PROJECT_STATUSES = 'statuses/'
+        i.e. https://na.myconnectwise.net becomes
+        https://api-na.myconnectwise.net
 
-    def get_project(self, project_id):
-        endpoint_url = '{}/{}'.format(self.ENDPOINT_PROJECTS, project_id)
-        return self.fetch_resource(endpoint_url)
+        See https://developer.connectwise.com/Products/Manage/Developer_Guide#Authentication  # noqa
+        """
+        url = urlparse(server_url)
+        if url.netloc not in CW_CLOUD_URLS:
+            # Don't change anything, just return.
+            return server_url
 
-    def get_projects(self, *args, **kwargs):
-        return self.fetch_resource(self.ENDPOINT_PROJECTS, should_page=True,
-                                   *args, **kwargs)
-
-    def get_project_statuses(self, *args, **kwargs):
-        return self.fetch_resource(self.ENDPOINT_PROJECT_STATUSES,
-                                   should_page=True,
-                                   *args, **kwargs)
+        return url._replace(netloc=CW_CLOUD_URLS[url.netloc]).geturl()
 
 
 class CompanyAPIClient(ConnectWiseAPIClient):
@@ -450,23 +479,60 @@ class ScheduleAPIClient(ConnectWiseAPIClient):
         endpoint_url = '{}/{}'.format(self.ENDPOINT_ENTRIES, entry_id)
         return self.fetch_resource(endpoint_url)
 
-    def post_schedule_entry(self, *args, **kwargs):
+    def post_schedule_entry(self, target, schedule_type, **kwargs):
         endpoint_url = self._endpoint(self.ENDPOINT_ENTRIES)
 
         body = {
-                    "objectId": kwargs.get("objectId"),
-                    "member": {
-                        "id": kwargs.get("resource").id,
-                        "identifier": kwargs.get("resource").identifier,
-                        "name": str(kwargs.get("resource")),
-                    },
+                    "objectId": target.id,
                     "type": {
-                        "id": kwargs.get("scheduleType").id,
-                        "identifier": kwargs.get("scheduleType").identifier,
+                        "id": schedule_type.id,
+                        "identifier": schedule_type.identifier,
                     }
                 }
+
+        member = kwargs.get("member")
+        if member:
+            body.update({
+                "member": {
+                            "id": member.id,
+                            "identifier": member.identifier,
+                            "name": str(member)
+                          }
+            })
+
+        date_start = kwargs.get("date_start")
+        if date_start:
+            body["dateStart"] = date_start.astimezone(
+                pytz.timezone('UTC')).strftime("%Y-%m-%dT%H:%M:%SZ")
+
+        date_end = kwargs.get("date_end")
+        if date_end:
+            body["dateEnd"] = kwargs.get("date_end")
+            body["dateEnd"] = date_end.astimezone(
+                pytz.timezone('UTC')).strftime("%Y-%m-%dT%H:%M:%SZ")
+
+        allow_conflicts = kwargs.get("allow_conflicts")
+        if allow_conflicts is not None:
+            body["allowScheduleConflictsFlag"] = allow_conflicts
+
         return self.request('post', endpoint_url, body)
 
+    def patch_schedule_entry(self, **kwargs):
+        schedule_id = kwargs.get("id")
+        endpoint_url = self._endpoint(
+            "{}/{}".format(self.ENDPOINT_ENTRIES, schedule_id))
+
+        # Yeah, this schema is a bit bizarre. See CW docs at
+        # https://developer.connectwise.com/Manage/Developer_Guide#Patch
+        body = [
+            {
+                'op': 'replace',
+                'path': 'doneFlag',
+                'value': kwargs.get("done_flag")
+            }
+        ]
+        return self.request('patch', endpoint_url, body)
+
     def delete_schedule_entry(self, entry_id):
         endpoint_url = self._endpoint(
             '{}/{}'.format(self.ENDPOINT_ENTRIES, entry_id))
@@ -475,6 +541,7 @@ class ScheduleAPIClient(ConnectWiseAPIClient):
             endpoint_url,
             auth=self.auth,
             timeout=self.timeout,
+            headers=self.get_headers(),
         )
 
     def get_calendars(self, *args, **kwargs):
@@ -506,12 +573,111 @@ class ScheduleAPIClient(ConnectWiseAPIClient):
 class TimeAPIClient(ConnectWiseAPIClient):
     API = 'time'
     ENDPOINT_ENTRIES = 'entries'
+    ENDPOINT_WORK_TYPES = 'workTypes'
+    ENDPOINT_WORK_ROLES = 'workRoles'
 
     def get_time_entries(self, *args, **kwargs):
         return self.fetch_resource(self.ENDPOINT_ENTRIES,
                                    should_page=True,
                                    *args, **kwargs)
 
+    def get_work_types(self, *args, **kwargs):
+        return self.fetch_resource(self.ENDPOINT_WORK_TYPES,
+                                   should_page=True,
+                                   *args, **kwargs)
+
+    def get_work_roles(self, *args, **kwargs):
+        return self.fetch_resource(self.ENDPOINT_WORK_ROLES,
+                                   should_page=True,
+                                   *args, **kwargs)
+
+    def post_time_entry(self, target_data, **kwargs):
+        endpoint_url = self._endpoint(self.ENDPOINT_ENTRIES)
+
+        time_start = kwargs.get("time_start")
+        time_start = time_start.astimezone(pytz.timezone('UTC')).strftime(
+            "%Y-%m-%dT%H:%M:%SZ")
+
+        body = {
+                    "chargeToId": target_data['id'],
+                    "chargeToType": target_data['type'],
+                    "timeStart": time_start,
+                    "addToDetailDescriptionFlag": kwargs
+                    .get("description_flag"),
+                    "addToInternalAnalysisFlag": kwargs
+                    .get("analysis_flag"),
+                    "addToResolutionFlag": kwargs.get("resolution_flag"),
+                    "emailResourceFlag": kwargs.get("resource_flag"),
+                    "emailContactFlag": kwargs.get("contact_flag"),
+                    "emailCcFlag": kwargs.get("cc_flag"),
+                }
+
+        member = kwargs.get("resource")
+        if member:
+            body.update({
+                "member": {
+                    "id": member.id,
+                    "identifier": member.identifier,
+                    "name": str(member),
+                }
+
+            })
+
+        time_end = kwargs.get("time_end")
+        if time_end:
+            time_end = time_end.astimezone(pytz.timezone('UTC')).strftime(
+                "%Y-%m-%dT%H:%M:%SZ")
+            body.update({"timeEnd": time_end})
+
+        hours_deduct = kwargs.get("hours_deduct")
+        if hours_deduct:
+            body.update({"hoursDeduct": hours_deduct})
+
+        actual_hours = kwargs.get("actual_hours")
+        if actual_hours:
+            body.update({"actualHours": actual_hours})
+
+        billable_option = kwargs.get("billable_option")
+        if billable_option:
+            body.update({"billableOption": billable_option})
+
+        work_type = kwargs.get("work_type")
+        if work_type:
+            body.update({
+                "workType": {
+                    "name": str(work_type)
+                }
+            })
+
+        work_role = kwargs.get("work_role")
+        if work_role:
+            body.update({
+                "workRole": {
+                    "name": str(work_role)
+                }
+            })
+
+        agreement = kwargs.get("agreement")
+        if agreement:
+            body.update({
+                "agreement": {
+                    "id": agreement.id,
+                    "name": str(agreement)
+                }
+            })
+
+        notes = kwargs.get("notes")
+        if notes:
+            body.update({"notes": notes})
+
+        email_cc = kwargs.get("email_cc")
+        if email_cc:
+            body.update({
+                "emailCc": email_cc
+            })
+
+        return self.request('post', endpoint_url, body)
+
 
 class SalesAPIClient(ConnectWiseAPIClient):
     API = 'sales'
@@ -521,6 +687,8 @@ class SalesAPIClient(ConnectWiseAPIClient):
     ENDPOINT_OPPORTUNITY_TYPES = \
         '{}/types'.format(ENDPOINT_OPPORTUNITIES)
     ENDPOINT_ACTIVITIES = 'activities'
+    ENDPOINT_ACTIVITY_STATUSES = '{}/statuses'.format(ENDPOINT_ACTIVITIES)
+    ENDPOINT_ACTIVITY_TYPES = '{}/types'.format(ENDPOINT_ACTIVITIES)
     ENDPOINT_PROBABILITIES = 'probabilities'
     ENDPOINT_STAGES = 'stages'
 
@@ -539,6 +707,16 @@ class SalesAPIClient(ConnectWiseAPIClient):
             self.ENDPOINT_ACTIVITIES, activity_id)
         return self.fetch_resource(endpoint_url)
 
+    def get_activity_statuses(self, *args, **kwargs):
+        return self.fetch_resource(self.ENDPOINT_ACTIVITY_STATUSES,
+                                   should_page=True,
+                                   *args, **kwargs)
+
+    def get_activity_types(self, *args, **kwargs):
+        return self.fetch_resource(self.ENDPOINT_ACTIVITY_TYPES,
+                                   should_page=True,
+                                   *args, **kwargs)
+
     def get_probabilities(self, *args, **kwargs):
         return self.fetch_resource(self.ENDPOINT_PROBABILITIES,
                                    should_page=True,
@@ -595,6 +773,27 @@ class SalesAPIClient(ConnectWiseAPIClient):
         ]
         return self.request('patch', endpoint_url, body)
 
+    def update_activity_status(self, obj_id, status):
+        """
+        Update the activities' status on the server.
+        """
+        # Yeah, this schema is a bit bizarre. See CW docs at
+        # https://developer.connectwise.com/Manage/Developer_Guide#Patch
+        endpoint_url = self._endpoint(
+            '{}/{}'.format(self.ENDPOINT_ACTIVITIES, obj_id)
+        )
+        body = [
+            {
+                'op': 'replace',
+                'path': 'status',
+                'value': {
+                    'id': status.id,
+                    'name': status.name,
+                },
+            },
+        ]
+        return self.request('patch', endpoint_url, body)
+
 
 class SystemAPIClient(ConnectWiseAPIClient):
     API = 'system'
@@ -667,6 +866,7 @@ class SystemAPIClient(ConnectWiseAPIClient):
                 endpoint,
                 auth=self.auth,
                 timeout=self.timeout,
+                headers=self.get_headers(),
             )
         except requests.RequestException as e:
             logger.error('Request failed: GET {}: {}'.format(endpoint, e))
@@ -701,13 +901,8 @@ class SystemAPIClient(ConnectWiseAPIClient):
         return m.group(1) if m else None
 
 
-class ServiceAPIClient(ConnectWiseAPIClient):
-    API = 'service'
+class TicketAPIMixin:
     ENDPOINT_TICKETS = 'tickets'
-    ENDPOINT_BOARDS = 'boards'
-    ENDPOINT_PRIORITIES = 'priorities'
-    ENDPOINT_LOCATIONS = 'locations'
-    ENDPOINT_SLAS = 'SLAs'
 
     def tickets_count(self):
         params = dict(
@@ -740,14 +935,6 @@ class ServiceAPIClient(ConnectWiseAPIClient):
                 'path': 'closedFlag',
                 'value': closed_flag
             },
-            {
-                'op': 'replace',
-                'path': 'priority',
-                'value': {
-                    'id': priority.id,
-                    'name': priority.name,
-                },
-            },
             {
                 'op': 'replace',
                 'path': 'status',
@@ -757,8 +944,28 @@ class ServiceAPIClient(ConnectWiseAPIClient):
                 },
             },
         ]
+
+        if priority:
+            priority_body = {
+                'op': 'replace',
+                'path': 'priority',
+                'value': {
+                    'id': priority.id,
+                    'name': priority.name,
+                },
+            }
+            body.append(priority_body)
+
         return self.request('patch', endpoint_url, body)
 
+
+class ServiceAPIClient(TicketAPIMixin, ConnectWiseAPIClient):
+    API = 'service'
+    ENDPOINT_BOARDS = 'boards'
+    ENDPOINT_PRIORITIES = 'priorities'
+    ENDPOINT_LOCATIONS = 'locations'
+    ENDPOINT_SLAS = 'SLAs'
+
     def get_notes(self, ticket_id, *args, **kwargs):
         """
         Returns the notes associated with the specific ticket.
@@ -767,6 +974,34 @@ class ServiceAPIClient(ConnectWiseAPIClient):
         return self.fetch_resource(endpoint_url, should_page=True,
                                    *args, **kwargs)
 
+    def post_note(self, target_data, **kwargs):
+        endpoint_url = self._endpoint(
+            '{}/{}/notes'.format(self.ENDPOINT_TICKETS, target_data['id']))
+
+        body = {
+            "detailDescriptionFlag": kwargs.get("description_flag"),
+            "internalAnalysisFlag": kwargs.get("analysis_flag"),
+            "resolutionFlag": kwargs.get("resolution_flag"),
+            "processNotifications": kwargs.get("process_notifications")
+        }
+
+        member = kwargs.get("resource")
+        if member:
+            body.update({
+                "member": {
+                    "id": member.id,
+                    "identifier": member.identifier,
+                    "name": str(member),
+                }
+
+            })
+
+        text = kwargs.get("text")
+        if text:
+            body.update({"text": text})
+
+        return self.request('post', endpoint_url, body)
+
     def get_statuses(self, board_id, *args, **kwargs):
         """
         Returns the status types associated with the specified board.
@@ -819,3 +1054,84 @@ class ServiceAPIClient(ConnectWiseAPIClient):
         endpoint_url = '{}/{}/items/'.format(self.ENDPOINT_BOARDS, board_id)
         return self.fetch_resource(endpoint_url, should_page=True,
                                    *args, **kwargs)
+
+
+class ProjectAPIClient(TicketAPIMixin, ConnectWiseAPIClient):
+    API = 'project'
+    ENDPOINT_PROJECTS = 'projects/'
+    ENDPOINT_PROJECT_STATUSES = 'statuses/'
+    ENDPOINT_PROJECT_PHASES = 'phases/'
+    ENDPOINT_PROJECT_TYPES = 'projectTypes/'
+
+    def get_project(self, project_id):
+        endpoint_url = '{}/{}'.format(self.ENDPOINT_PROJECTS, project_id)
+        return self.fetch_resource(endpoint_url)
+
+    def get_projects(self, *args, **kwargs):
+        return self.fetch_resource(self.ENDPOINT_PROJECTS, should_page=True,
+                                   *args, **kwargs)
+
+    def get_project_statuses(self, *args, **kwargs):
+        return self.fetch_resource(self.ENDPOINT_PROJECT_STATUSES,
+                                   should_page=True,
+                                   *args, **kwargs)
+
+    def get_project_types(self, *args, **kwargs):
+        return self.fetch_resource(self.ENDPOINT_PROJECT_TYPES,
+                                   should_page=True,
+                                   *args, **kwargs)
+
+    def get_project_phases(self, project_id, *args, **kwargs):
+        endpoint_url = '{}{}/{}'.format(
+            self.ENDPOINT_PROJECTS, project_id, self.ENDPOINT_PROJECT_PHASES
+        )
+        return self.fetch_resource(endpoint_url,
+                                   should_page=True,
+                                   *args, **kwargs)
+
+
+class FinanceAPIClient(ConnectWiseAPIClient):
+    API = 'finance'
+    ENDPOINT_AGREEMENTS = 'agreements'
+
+    def get_agreements(self, *args, **kwargs):
+        return self.fetch_resource(self.ENDPOINT_AGREEMENTS, should_page=True,
+                                   *args, **kwargs)
+
+
+class HostedAPIClient(SystemAPIClient):
+    ENDPOINT_HOSTED_SETUPS = 'connectwisehostedsetups/'
+
+    def post_hosted_tab(self, *args, **kwargs):
+        endpoint_url = self._endpoint(self.ENDPOINT_HOSTED_SETUPS)
+
+        body = {
+            'screenId': kwargs.get('screen_id'),
+            'description': kwargs.get('description'),
+            'url': kwargs.get('url'),
+            'origin': kwargs.get('origin'),
+            'type': kwargs.get('type')
+        }
+
+        return self.request('post', endpoint_url, body)
+
+    def get_hosted_tabs(self, *args, **kwargs):
+        return self.fetch_resource(self.ENDPOINT_HOSTED_SETUPS,
+                                   should_page=True,
+                                   *args, **kwargs)
+
+    def delete_hosted_tab(self, hosted_tab_id):
+        endpoint_url = self._endpoint(
+            '{}/{}'.format(self.ENDPOINT_HOSTED_SETUPS, hosted_tab_id)
+        )
+
+        return requests.request('delete', endpoint_url, auth=self.auth)
+
+
+class HostedReportAPIClient(ConnectWiseAPIClient):
+    API = 'system/reports'
+    ENDPOINT_HOSTED_REPORT = 'ConnectWiseHostedApiScreen/'
+
+    def get_hosted_screen_ids(self, *args, **kwargs):
+        return self.fetch_resource(self.ENDPOINT_HOSTED_REPORT,
+                                   *args, **kwargs)
