@@ -1,5 +1,4 @@
 import io
-import unittest
 
 from django.core.management import call_command
 from django.test import TestCase
@@ -38,7 +37,6 @@ class AbstractBaseSyncTest(object):
         args = ['cwsync', cw_object]
         if full_option:
             args.append('--full')
-
         call_command(*args, stdout=out)
         return out
 
@@ -185,6 +183,14 @@ class TestSyncProjectStatusesCommand(AbstractBaseSyncTest, TestCase):
     )
 
 
+class TestSyncProjectTypesCommand(AbstractBaseSyncTest, TestCase):
+    args = (
+        mocks.projects_api_get_project_types_call,
+        fixtures.API_PROJECT_TYPES,
+        'project_type',
+    )
+
+
 class TestSyncProjectsCommand(AbstractBaseSyncTest, TestCase):
     args = (
         mocks.project_api_get_projects_call,
@@ -197,6 +203,19 @@ class TestSyncProjectsCommand(AbstractBaseSyncTest, TestCase):
         fixture_utils.init_project_statuses()
 
 
+class TestSyncProjectPhaseCommand(AbstractBaseSyncTest, TestCase):
+    args = (
+        mocks.projects_api_get_project_phases_call,
+        fixtures.API_PROJECT_PHASE_LIST,
+        'project_phase'
+    )
+
+    def setUp(self):
+        super().setUp()
+        fixture_utils.init_projects()
+        fixture_utils.init_project_phases()
+
+
 class TestSyncBoardsStatusesCommand(AbstractBaseSyncTest, TestCase):
     args = (
         mocks.service_api_get_statuses_call,
@@ -375,7 +394,22 @@ class TestSyncScheduleStatusesCommand(AbstractBaseSyncTest, TestCase):
     )
 
 
-@unittest.skip("Activities sync is temporarily removed- see issue #484")
+class TestSyncActivityStatusesCommand(AbstractBaseSyncTest, TestCase):
+    args = (
+        mocks.sales_api_get_activities_statuses_call,
+        fixtures.API_SALES_ACTIVITY_STATUSES,
+        'activity_status',
+    )
+
+
+class TestSyncActivityTypesCommand(AbstractBaseSyncTest, TestCase):
+    args = (
+        mocks.sales_api_get_activities_types_call,
+        fixtures.API_SALES_ACTIVITY_TYPES,
+        'activity_type',
+    )
+
+
 class TestSyncActivityCommand(AbstractBaseSyncTest, TestCase):
     args = (
         mocks.sales_api_get_activities_call,
@@ -438,6 +472,58 @@ class TestSyncItemCommand(AbstractBaseSyncTest, TestCase):
     )
 
 
+class TestSyncWorkTypeCommand(AbstractBaseSyncTest, TestCase):
+    def setUp(self):
+        super().setUp()
+        fixture_utils.init_work_types()
+
+    args = (
+        mocks.time_api_get_work_types_call,
+        fixtures.API_WORK_TYPE_LIST,
+        'work_type'
+    )
+
+
+class TestSyncWorkRoleCommand(AbstractBaseSyncTest, TestCase):
+    def setUp(self):
+        super().setUp()
+        fixture_utils.init_work_roles()
+
+    args = (
+        mocks.time_api_get_work_roles_call,
+        fixtures.API_WORK_ROLE_LIST,
+        'work_role'
+    )
+
+
+class TestSyncAgreementCommand(AbstractBaseSyncTest, TestCase):
+    def setUp(self):
+        super().setUp()
+        fixture_utils.init_agreements()
+
+    args = (
+        mocks.finance_api_get_agreements_call,
+        fixtures.API_AGREEMENT_LIST,
+        'agreement'
+    )
+
+
+class TestSyncProjectTicketCommand(AbstractBaseSyncTest, TestCase):
+    def setUp(self):
+        super().setUp()
+        mocks.system_api_get_member_image_by_photo_id_call(
+            (mocks.CW_MEMBER_IMAGE_FILENAME, mocks.get_member_avatar()))
+        fixture_utils.init_members()
+        fixture_utils.init_board_statuses()
+        fixture_utils.init_project_tickets()
+
+    args = (
+        mocks.project_api_tickets_test_command,
+        fixtures.API_PROJECT_TICKET_LIST,
+        'project_ticket'
+    )
+
+
 class TestSyncAllCommand(TestCase):
 
     def setUp(self):
@@ -453,6 +539,7 @@ class TestSyncAllCommand(TestCase):
             TestSyncCompanyTypesCommand,
             TestSyncLocationsCommand,
             TestSyncPrioritiesCommand,
+            TestSyncProjectPhaseCommand,
             TestSyncProjectStatusesCommand,
             TestSyncProjectsCommand,
             TestSyncTeamsCommand,
@@ -465,7 +552,9 @@ class TestSyncAllCommand(TestCase):
             TestSyncOpportunityTypesCommand,
             TestSyncOpportunityCommand,
             TestSyncSalesProbabilitiesCommand,
-            # TestSyncActivityCommand,
+            TestSyncActivityStatusesCommand,
+            TestSyncActivityTypesCommand,
+            TestSyncActivityCommand,
             TestSyncScheduleTypesCommand,
             TestSyncScheduleStatusesCommand,
             TestSyncScheduleEntriesCommand,
@@ -479,7 +568,11 @@ class TestSyncAllCommand(TestCase):
             TestSyncHolidayListCommand,
             TestSyncTypeCommand,
             TestSyncSubTypeCommand,
-            TestSyncItemCommand
+            TestSyncItemCommand,
+            TestSyncWorkTypeCommand,
+            TestSyncWorkRoleCommand,
+            TestSyncAgreementCommand,
+            TestSyncProjectTypesCommand
         ]
 
         self.test_args = []
@@ -489,7 +582,7 @@ class TestSyncAllCommand(TestCase):
             apicall, fixture, cw_object = test_case.args
             apicall(fixture)
 
-    def _test_sync(self, full_option=False):
+    def run_sync_command(self, full_option=False):
         out = io.StringIO()
         args = ['cwsync']
 
@@ -500,10 +593,13 @@ class TestSyncAllCommand(TestCase):
 
         return out.getvalue().strip()
 
-    def test_sync(self):
-        """Test sync all objects command."""
+    def test_partial_sync(self):
+        """
+        Test the command to run a sync of all objects without
+        the --full argument.
+        """
+        output = self.run_sync_command()
 
-        output = self._test_sync()
         for apicall, fixture, cw_object in self.test_args:
             summary = sync_summary(slug_to_title(cw_object), len(fixture))
             self.assertIn(summary, output)
@@ -526,13 +622,14 @@ class TestSyncAllCommand(TestCase):
                          len([fixtures.API_SERVICE_TICKET]))
 
     def test_full_sync(self):
-        """Test sync all objects command."""
+        """Test the command to run a full sync of all objects."""
         cw_object_map = {
             'member': models.Member,
             'board': models.ConnectWiseBoard,
             'priority': models.TicketPriority,
             'project_status': models.ProjectStatus,
             'project': models.Project,
+            'project_phase': models.ProjectPhase,
             'board_status': models.BoardStatus,
             'territory': models.Territory,
             'company_status': models.CompanyStatus,
@@ -548,7 +645,9 @@ class TestSyncAllCommand(TestCase):
             'opportunity_stage': models.OpportunityStage,
             'opportunity_type': models.OpportunityType,
             'sales_probability': models.SalesProbability,
-            # 'activity': models.Activity,
+            'activity_status': models.ActivityStatus,
+            'activity_type': models.ActivityType,
+            'activity': models.Activity,
             'schedule_entry': models.ScheduleEntry,
             'schedule_type': models.ScheduleType,
             'schedule_status': models.ScheduleStatus,
@@ -561,19 +660,25 @@ class TestSyncAllCommand(TestCase):
             'holiday_list': models.HolidayList,
             'type': models.Type,
             'sub_type': models.SubType,
-            'item': models.Item
+            'item': models.Item,
+            'work_type': models.WorkType,
+            'work_role': models.WorkRole,
+            'agreement': models.Agreement,
+            'project_type': models.ProjectType,
         }
 
-        self.test_sync()
+        # Run partial sync first
+        self.run_sync_command()
 
+        fixture_utils.init_projects()
         fixture_utils.init_members()
         fixture_utils.init_board_statuses()
         fixture_utils.init_teams()
 
         pre_full_sync_counts = {}
 
-        for key, clazz in cw_object_map.items():
-            pre_full_sync_counts[key] = clazz.objects.all().count()
+        for key, model_class in cw_object_map.items():
+            pre_full_sync_counts[key] = model_class.objects.all().count()
 
         mocks.system_api_get_members_call([])
         mocks.company_api_by_id_call([])
@@ -581,19 +686,34 @@ class TestSyncAllCommand(TestCase):
         for apicall, _, _ in self.test_args:
             apicall([])
 
-        output = self._test_sync(full_option=True)
+        output = self.run_sync_command(full_option=True)
 
         for apicall, fixture, cw_object in self.test_args:
-            summary = full_sync_summary(slug_to_title(cw_object),
-                                        pre_full_sync_counts[cw_object])
+            if cw_object in (
+                    'team',
+                    'board_status',
+                    'opportunity_note',
+                    'sla_priority',
+                    'holiday',
+            ):
+                # Assert that there were objects to get deleted, then change
+                # to zero to verify the output formats correctly.
+                # We are just testing the command, there are sync tests to
+                # verify that the syncronizers work correctly
+                self.assertGreater(pre_full_sync_counts[cw_object], 0)
+                pre_full_sync_counts[cw_object] = 0
+            summary = full_sync_summary(
+                slug_to_title(cw_object),
+                pre_full_sync_counts[cw_object]
+            )
             self.assertIn(summary, output)
 
 
-class TestListMembersCommand(TestCase):
+class TestListUsersCommand(TestCase):
     def test_command(self):
         # We don't need to check output carefully. Just verify it
         # doesn't explode.
         self.assertEqual(
-            call_command('list_members'),
+            call_command('list_users'),
             None
         )
